package common;

import org.apache.http.HttpEntity;
import org.apache.http.HttpStatus;
import org.apache.http.NameValuePair;
import org.apache.http.client.config.RequestConfig;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.client.utils.URIBuilder;
import org.apache.http.entity.StringEntity;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClientBuilder;
import org.apache.http.message.BasicNameValuePair;
import org.apache.http.util.EntityUtils;

import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * 全局的http客户端请求工具
 * <p>
 * 文档表明该类 {@link CloseableHttpClient} 是线程安全的类, 也是个很重的类, 不适宜每次请求都进行初始化
 * 默认会初始化一个客户端连接池, 但是需要注意的是默认为每个域名分配的最大连接数是5, 连接池最大连接数是2*5=10
 * 所有http请求共享一个客户端连接池恐怕峰值阶段不够用, 需要手动设置为偏大的连接池
 * http请求相对比较慢, 会阻塞其它请求, 多不是瞬时的请求
 * <p>
 * 该配置在一个jvm进程中生效, 故太多的请求会因其它负载(比如网关/nginx等)分流, 无需给同一个jvm进程分配过多的线程数
 */
public class HttpUtils {

    private HttpUtils() {
    }

    private static final CloseableHttpClient GLOBAL_HTTP_CLIENT;

    static {
        // 存在连接超时和数据超时, 会导致请求卡住, 需要主动放弃
        RequestConfig config = RequestConfig.custom()
                // 设置连接超时时间, 单位毫秒
                .setConnectTimeout(2000)
                // 请求获取数据的超时时间, 单位毫秒
                // 如果访问一个接口, 多少时间内无法返回数据, 就直接放弃此次调用
                .setSocketTimeout(10000)
                .build();
        GLOBAL_HTTP_CLIENT = HttpClientBuilder.create()
                // 全局请求超时策略
                .setDefaultRequestConfig(config)
                // 客户端连接池数量, 太多线程会导致线程上下文切换频繁, 开销也相当大
                .setMaxConnTotal(100)
                // 每个域名可占用的连接池数量, 考虑到订单详情访问的比较多, 超过会阻塞
                .setMaxConnPerRoute(30)
                .build();
    }

    /**
     * 发起HTTP GET请求
     *
     * @param url    请求地址, 必须存在
     * @param params 请求参数内容, 可能不存在
     * @return byte array containing the entity content of response
     * @throws Exception 请求过程中发生错误
     */
    public static byte[] get(String url, Map<String, String> params) throws Exception {
        URI uri;
        try {
            URIBuilder builder = new URIBuilder(url);
            if (params != null) {
                for (String key : params.keySet()) {
                    builder.addParameter(key, params.get(key));
                }
            }
            uri = builder.build();
        } catch (URISyntaxException e) {
            throw new Exception("url地址错误, url: " + url);
        }

        final HttpGet httpGet = new HttpGet(uri);
        try (final CloseableHttpResponse response = GLOBAL_HTTP_CLIENT.execute(httpGet)) {
            if (response.getStatusLine().getStatusCode() != HttpStatus.SC_OK) {
                throw new Exception("请求url地址时发生错误, url: " + url);
            }
            return EntityUtils.toByteArray(response.getEntity());
        } catch (IOException e) {
            throw new Exception("对url地址返回的数据进行转换时发生错误, url: " + url);
        }
    }

    /**
     * 发起HTTP GET请求
     *
     * @param url 请求地址, 必须存在
     * @return byte array containing the entity content of response
     */
    public static byte[] get(String url) throws Exception {
        return get(url, null);
    }

    /**
     * application/json 方式请求
     *
     * @param host    请求地址
     * @param headers 请求头, 使用默认时传 null
     * @param params  请求参数, 无时传 null, 就是拼接在主机名后的key=value参数, 可直接处理成host代替
     * @param body    请求体
     * @return {@link String} response
     */
    public static String postJson(String host, Map<String, String> headers, Map<String, String> params, String body) {
        return post(host, getHeaders(headers, RequestContentType.JSON), params, RequestContentType.JSON, body);
    }

    /**
     * application/x-www-form-urlencoded 方式请求
     *
     * @param host    请求地址
     * @param headers 请求头, 使用默认传 null
     * @param params  请求参数
     * @param body    如果 uri 为无 query 查询参数, 则需要按 body 的方式传入, 无时默认为 null 即可
     * @return {@link String} response
     */
    public static String postUrlEncoded(String host, Map<String, String> headers, Map<String, String> params, String body) {
        return post(host, getHeaders(headers, RequestContentType.URL_ENCODED), params, RequestContentType.URL_ENCODED, body);
    }

    /**
     * 获取头信息, 无时提供内容的头信息
     *
     * @param headers 已有头信息
     * @param type    头内容类型
     * @return 头内容类型
     */
    private static Map<String, String> getHeaders(Map<String, String> headers, RequestContentType type) {
        if (headers != null && headers.size() >= 1) {
            return headers;
        }
        Map<String, String> defaultHeaders = new HashMap<>();
        switch (type) {
            case JSON:
                defaultHeaders.put("Content-Type", "application/json");
                break;
            case URL_ENCODED:
                defaultHeaders.put("Content-Type", "application/x-www-form-urlencoded");
                break;
            default:
        }
        return defaultHeaders;
    }

    /**
     * 请求头内容类型
     */
    private enum RequestContentType {
        // Content-Type: application/json
        JSON,
        // Content-Type: application/x-www-form-urlencoded
        URL_ENCODED
    }

    /**
     * 发起 http post请求
     *
     * @param uri     请求主机名称, 必须存在
     * @param headers 请求头信息, 内部已处理
     * @param params  请求参数内容, 可能不存在
     * @param type    请求头内容类型
     * @param body    请求内容体, 可能不存在
     * @return 响应
     */
    private static String post(
            String uri,
            Map<String, String> headers,
            Map<String, String> params,
            RequestContentType type,
            String body
    ) {

        HttpPost httpPost = null;
        CloseableHttpResponse response = null;

        // 添加参数
        List<NameValuePair> pairList = new ArrayList<>();
        if (params != null && params.size() > 0) {
            for (Map.Entry<String, String> entry : params.entrySet()) {
                pairList.add(new BasicNameValuePair(entry.getKey(), entry.getValue()));
            }
        }

        try {
            // 请求参数除body外最终都需要拼接到地址栏中
            // 如果需要单独设置 Entity 再处理
            URIBuilder builder = new URIBuilder(uri);
            if (pairList.size() > 0) {
                builder.setParameters(pairList);
            }
            httpPost = new HttpPost(builder.build());

            // entity 实体处理
            // 对实体的理解, 看实现方式的不同而不同
            // 如果 uri 后面没有查询 query 参数, 则需要按 body 的方式传入参数
            // 故该处无法区分是否仅 json 格式时需要 body 参数
            HttpEntity entity = null;
            // 有 body 内容需要处理body
            if (body != null && body.length() > 0) {
                entity = new StringEntity(body, StandardCharsets.UTF_8);
            }

            if (entity != null) {
                httpPost.setEntity(entity);
            }

            // 添加头内容
            for (Map.Entry<String, String> entry : headers.entrySet()) {
                httpPost.setHeader(entry.getKey(), entry.getValue());
            }

            response = GLOBAL_HTTP_CLIENT.execute(httpPost);
            int code = response.getStatusLine().getStatusCode();
            String content = EntityUtils.toString(response.getEntity(), StandardCharsets.UTF_8.name());
            if (code == HttpStatus.SC_OK) {
                return content;
            }
            throw new RuntimeException("服务繁忙");
        } catch (Exception e) {
            throw new RuntimeException(e);
        } finally {
            if (response != null) {
                try {
                    response.close();
                } catch (Exception e) {
                }
            }
            if (httpPost != null) {
                httpPost.releaseConnection();
            }
        }
    }
}
